/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef BITFIELDBASE_H
#define BITFIELDBASE_H

#include <string>

#include <bit/enums.h>
#include <bit/pointer.h>
#include <bit/except.h>
#include <bit/fieldtype.h>

namespace bit
  {

    class RecordBase;
    class RecordStorage;

  /**
   * FieldBase serves as a top level parent providing the concepts of a starting
   * offset within a buffer, a length value, and units associated with each.
   *
   * @todo Put in some type of check to ensure validity of a set type, or defer until access???
   * @ingroup records
   * @author Rick L. Vinyard, Jr.
   */
  class FieldBase
    {
    public:

      /** A convenience typedef to a smart pointer to FieldBase */
      typedef BitPointer<FieldBase> pointer;

      /**
       * Default constructor
       */
      FieldBase();

      /** Destructor */
      virtual ~FieldBase();

      /** @return the length of this object in it's length units */
      virtual size_t length() const = 0;

      /**
       * @return the length of this object in the supplied units
       *
       * Whether the object is specified in bits or octets, length(BITS) will always
       * return the length in bits and length(OCTETS) will always return the
       * length in octets.
       *
       * When the length in bits is not evenly divisible by the requested unit size
       * (say a 12 bit offset with requested units = 8) the return value will
       * be the number of units required to store the object.
       * For example, if the object has a 12 bit offset and the units parameter = OCTETS,
       * the return value will be 2.
       */
      virtual size_t length(size_t units) const;

      /** Returns the length units */
      virtual size_t length_units() const = 0;

      /** @return the offset of this object in the offset units of this object */
      virtual int offset() const = 0;

      /**
       * @return the offset of this object in the specified offset units
       *
       * This method can be used to get the offset as a specified unit size
       * regardless of the currently set offset unit size of this object.
       *
       * When the offset size in bits is not evenly divisible by the requested unit size
       * (say a 12 bit offset with requested units = 8) the return value will
       * be the number of units required to complete the offset.
       * For example, if the object has a 12 bit offset and the units parameter = 8,
       * the return value will be 2. If the units were the same, but the offset were -12
       * bits the return value would be -2.
       */
      virtual int offset(size_t units) const;

      /** Returns the offset units */
      virtual size_t offset_units() const = 0;

      /**
       * @return the starting position of this object in the specified units
       *
       * This method will return the starting position of this object in the specified
       * units, which defaults to bits. The returned value will also account for the
       * offset of the parent object (if this object has a parent).
       *
       * When the offset size in bits is not evenly divisible by the requested unit size
       * (say a 12 bit offset with requested units = 8) the return value will
       * be the number of units required to complete the offset.
       * For example, if the object has a 12 bit offset and the units parameter = 8,
       * the return value will be 2. If the units were the same, but the offset were -12
       * bits the return value would be -2.
       */
      virtual int start(size_t units=BITS) const;

      /**
       * Returns the string name associated with this object
       *
       * This pure virtual method is reimplemented by children to
       * provide the proper notation, such as the array notation of
       * elements, by children.
       */
      virtual std::string name() const = 0;

      virtual std::string name(int depth) const;

      virtual size_t depth() const;

      /**
       * Returns the string description of this object.
       */
      virtual std::string description() const = 0;

      /**
       * Returns a FieldType object that describes how the field data should be interpreted.
       */
      const FieldType& type();

      /**
       * Sets the FieldType object describing how the field data should be interpreted.
       */
      void set_type(const FieldType&);

      virtual const std::string& xml() = 0;

      /**
       * Compares the starting bit position of objects.
       *
       * This overload is provided to allow easy sorting of objects according to their
       * starting bit position.
       *
       * @return true if this object's starting bit is less than the other object's starting bit
       */
      virtual bool operator<(const FieldBase& other) const;

      /**
       * Compares the ending bit position of objects.
       *
       * This overload is provided to allow easy sorting of objects according to position
       * of their last bit.
       *
       * @return true if this object's starting bit is less than the other object's starting bit
       */
      virtual bool operator>(const FieldBase& other) const;

      virtual FieldBase::pointer clone() = 0;

      /**
       * This is a generalized interator that performs all operations using the generalized container methods
       * in the FieldBase class.
       */
    class iterator: public std::iterator<std::bidirectional_iterator_tag, FieldBase>
        {
        public:
          iterator(FieldBase* container=NULL, FieldBase::pointer object=FieldBase::pointer());

          ~iterator();

          iterator& operator=(const iterator& other);

          bool operator==(const iterator& other) const;

          bool operator!=(const iterator& other) const;

          iterator& operator++() throw (exception::invalid_iterator);

          iterator operator++(int) throw (exception::invalid_iterator);

          iterator& operator--() throw (exception::invalid_iterator);

          iterator operator--(int) throw (exception::invalid_iterator);

          FieldBase& operator*() throw (exception::invalid_iterator);

          FieldBase* operator->() throw (exception::invalid_iterator);

          FieldBase::pointer pointer();

        private:
          FieldBase* m_container;
          FieldBase::pointer m_object;
        };

      virtual iterator begin();

      virtual iterator end();

      virtual FieldBase& operator[](std::string s) throw (exception::invalid_container_op, std::out_of_range);

      virtual FieldBase& operator[](size_t i) throw (exception::invalid_container_op, std::out_of_range);

      virtual size_t fields();

      virtual FieldBase::pointer field(std::string s);

      virtual FieldBase::pointer field(size_t i);

      FieldBase* parent() { return m_parent; }

      virtual Container container_type() = 0;

    protected:
      friend class RecordStorage;
      friend class RecordVector;

      FieldBase* m_parent;
      FieldType m_type;
      std::string m_xml;

      virtual FieldBase::pointer previous_field(FieldBase::pointer current_field) throw (exception::invalid_container_op);

      virtual FieldBase::pointer next_field(FieldBase::pointer current_field) throw (exception::invalid_container_op);

    };

}

#endif
